package edu.assistment.experiments.inclinedplane.model;

import edu.assistment.simulation.model.Simulation;

/**
 * Model behind the inclined plane.
 * 
 * y
 * ^
 * |__
 * |  \
 * |   \            __/
 * |    \        __/
 * |     \    __/
 * |      \__/
 * --------------------------------> x
 * 
 *  Note that the axes represented by the model are
 *  NOT the same as the view coordinates on the screen.
 *  
 *  The model assumes that the ball rolls horizontally at
 *  y = 0.
 * 
 * @author Mike Sao Pedro (mikesp@wpi.edu)
 */
public class InclinedPlaneModel extends Simulation
{

	// --- STATIC PARTS OF THE MODEL ---
	
	/**
	 * Compute model changes at this dt
	 */
	private double dt = 0.1;
	
	/**
	 * Properties of the incline
	 */
	private double incline_theta_rad = Math.PI/4;
	
	/**
	 * What % of the way the ball is up the incline
	 */
	private double incline_ball_start_percent = 0.50;
	private double incline_length = 250;
	private double incline_friction = .1;
	
	/**
	 * Properties of the ball
	 */
	private double ball_mass = 1;
	private double ball_radius = 10;
	
	/**
	 * Properties of the run length
	 */
	private double run_length_theta = 0.0;
	private double run_length = 500;
	private double run_length_friction = .1;
	
	// --- DYNAMIC PARTS OF THE MODEL
	private double ball_x;
	private double ball_y;
	private double ball_speed;

//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
	
	/**
	 * Use the set methods to change values of the model.
	 */
	public InclinedPlaneModel()
	{
		super();
		ball_speed = 0.0;
		setBallAgainstRamp();
	}
	
	public double getDt() { return dt; }
	public double getTheta() { return incline_theta_rad; }
	public double getInclineFriction() { return incline_friction; }
	public double getBallMass() { return ball_mass; }
	public double getRunLength() { return run_length; }
	public double getRunIncline() { return run_length_theta; }
	public double getRunLengthFriction() { return run_length_friction; }
	public double getBallX() { return ball_x; }
	public double getBallY() { return ball_y; }
	public double getBallSpeed() { return ball_speed; }
	public double getInclineLength() { return incline_length; }
	public double getInclineHeight() { return Math.sin(getTheta()); }
	public double getBallStartPercent() { return incline_ball_start_percent; }
	public double getBallRadius() { return ball_radius; }
	public double getXIntercept() { return getInclineLength()*Math.cos(getTheta()); }
	public double getYIntercept() { return getInclineLength()*Math.sin(getTheta()); }
	public double getDistanceRolledFromIncline()
	{
		double x_intercept = getXIntercept();
		return Math.round(getBallX() < x_intercept ? 0.0 : (getBallX() - getXIntercept())); 
	}
	
	public void setDt(double val) { dt = val; }
	
	/**
	 * @param val in RADIANS
	 */
	public void setInclineTheta(double val) { incline_theta_rad = val; }
	public void setInclineFriction(double val) { incline_friction = val; }
	public void setBallMass(double val) { ball_mass = val; }
	public void setRunLength(double val) { run_length = val; }
	public void setRunTheta(double val) { run_length_theta = val; }
	public void setRunLengthFriction(double val) { run_length_friction = val; }
	public void setBallX(double val) { ball_x = val; }
	public void setBallY(double val) { ball_y = val; }
	public void setBallSpeed(double val) { ball_speed = val; }
	public void setInclineLength(double val) { incline_length = val; }
	public void setBallStartPercent(double val) { incline_ball_start_percent = val; }
	public void setBallRadius(double radius) { ball_radius = radius; }
	
	/**
	 * Resets the balls CENTER against the ramp, at the bottom of the incline.
	 */
	public void setBallAgainstRamp()
	{
		// --- Compute ramp x-and-y-intercept (y is 0 at this point)
		// --- Now we have a vector. Make it unit length and translate
		//     that many % up the incline
		//
	    double x_intercept = getXIntercept();
	    
	    // --- The point
	    double tangent_pt_x = x_intercept - x_intercept*getBallStartPercent();
	    double tangent_pt_y = getYIntercept()*getBallStartPercent();

	    setBallX((int)(tangent_pt_x + getBallRadius()*Math.sin(getTheta())));
	    setBallY((int)(tangent_pt_y + getBallRadius()*Math.cos(getTheta())));
	    setBallSpeed(0.0);
	}

//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
	
	public String modelStats()
	{
		return "<< TIME = " + getCurrentTime() + " >>\n" +
		       "Xpos: " + getBallX() + "\n" +
		       "Ypos: " + getBallY() + "\n" +
		       "Speed: " + getBallSpeed() + "\n" +
		       "Length: " + getRunLength() + "\n" +
		       "*************************************\n";
	}
	
}