package edu.assistment.experiments.inclinedplane.simulation;

import edu.assistment.experiments.inclinedplane.model.InclinedPlaneModel;
import edu.assistment.simulation.model.SimulationEngine;

/**
 * Update the ball's position for the next dt using the
 * given model.
 * 
 * @author Mike Sao Pedro (mikesp@wpi.edu)
 *
 */
public class InclinedPlaneSimulationEngine extends SimulationEngine
{
	private static final double GRAVITY = 9.8;
	
	private InclinedPlaneModel model = null;
	
	public InclinedPlaneSimulationEngine(InclinedPlaneModel model)
	{
		super();
		this.model = model;
	}
	
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
	
	/**
	 * Compute the next step
	 * 
	 * @return true if there is a next step to be computed; false if last step completed
	 */
	public boolean nextStep()
	{
		//System.err.println("NEXT STEP!!!");
		
		// --- If the ball is on the runway, do something different
		if(Double.compare(model.getBallY(), model.getBallRadius()) <= 0 ||
		   Double.compare(model.getDistanceRolledFromIncline(), 0.0) > 0
		   )
		{
			// --- If the ball's speed is zero, we're DONE
			if(Double.compare(model.getBallSpeed(), 0.0) <= 0)
				return false;
			
			updateOnRunway();
		}
		else
		{
			updateOnRamp();
		}
		
		model.setCurrentTime(model.getCurrentTime() + model.getDt());
		//System.err.println(model.modelStats());
		return true;
	}

	
	private void updateOnRamp()
	{
		// --- Compute Angle Movement
		double x_delta = Math.cos(model.getTheta());
		double y_delta = Math.sin(model.getTheta());
		
		// --- Increase Speed
		model.setBallSpeed(
				model.getBallSpeed() + 
				(GRAVITY*y_delta*model.getBallMass()*model.getDt() -
			    model.getInclineFriction()));

		// --- Change position
		//     Since the ball is rolling down an incline, y should DECREASE
		model.setBallX(model.getBallX() + (model.getBallSpeed()*x_delta));
		model.setBallY(model.getBallY() - (model.getBallSpeed()*y_delta));
		
		
		// --- Ensure we don't overrun the ramp part
		if(Double.compare(model.getBallY(), model.getBallRadius()) <= 0)
		{
			model.setBallY(model.getBallRadius());			
		}
	}
	
	
	private void updateOnRunway()
	{

		// --- Compute Angle Movement
		double x_delta = Math.cos(model.getRunIncline());
		double y_delta = Math.sin(model.getRunIncline());
		double decrease_factor = Math.random()*0.25;

		// --- Decrease Speed
		// --- Decrease ball speed by the friction coeff + ball resistance
		//     Hokey equation, but that's ok. We don't need anything special
		//
		//     Add a random element for some variation
		model.setBallSpeed(
				 model.getBallSpeed() - 
				 (GRAVITY*y_delta*model.getBallMass()*model.getDt()) -
				 decrease_factor - 
				 (model.getRunLengthFriction())
				);

		// --- Change position
		//     Since the ball is rolling up an incline, y should INCREASE
		model.setBallX(model.getBallX() + (model.getBallSpeed()*x_delta));
		model.setBallY(model.getBallY() + (model.getBallSpeed()*y_delta));
				
		// --- Ensure we don't overrun the ramp part (start rolling backwards)
		if(Double.compare(model.getBallSpeed(), 0.0) < 0)
		{
			model.setBallSpeed(0.0);			
		}

	}
}
