package edu.assistment.experiments.inclinedplane.view;

import java.awt.*;

import edu.assistment.experiments.inclinedplane.model.InclinedPlaneModel;
import edu.assistment.simulation.model.ModelChangeListener;

/**
 * Canvas for drawing an inclined plane and ball
 * rolling down it
 * 
 * @author Mike Sao Pedro (mikesp@wpi.edu)
 */
public class InclinedPlaneCanvas extends Canvas implements ModelChangeListener
{

	/**
	 * 
	 */
	private static final long serialVersionUID = 1L;
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////

	private InclinedPlaneModel model;
	
	// For Double Buffering
	private Graphics offScreenGraphics;
	private Image offScreenImage;
	
	/**
	 * Only change THIS value to change the size of the ball
	 */
	private final static int INCLINE_HORIZONTAL_OFFSET = 20;
	private final static int INCLINE_VERTICAL_OFFSET = 10;
	
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
	
	/**
	 * Create a new InclinedPlaneModel and set this object as a listener
	 * to model changes.
	 */
	public InclinedPlaneCanvas()
	{
		this(new InclinedPlaneModel());
		restart();
		repaint();
	}
	
	public InclinedPlaneCanvas(InclinedPlaneModel m) 
	{
		super();
		model = m;
		model.addModelChangeListener(this);
		this.setPreferredSize(new Dimension(300,200));
		restart();
		repaint();
	}

	public void restart()
	{
		offScreenImage = null;
	}
	
	public InclinedPlaneModel getInclinedPlaneModel()
	{
		return model;
	}
	
	public Dimension minimumSize()
	{
		return new Dimension(300,200);
	}
	
	public Dimension getPreferredSize()
	{
		return new Dimension(300,200);
	}

	public void clearCanvas()
	{
	    if(offScreenImage != null) 
	    {
	      Graphics osg = offScreenImage.getGraphics();
	      osg.setColor(getBackground());
	      osg.fill3DRect(0, 0, size().width, size().height, false);
	    }	
	}
	
//	  //===================================================================
//	  // increment
//	  //
//	  // Increment the double well oscillator, map the new point onto the
//	  // window.
//	  //===================================================================
//	  public void increment() {
//
//	    oldWindowxPos = windowxPos;
//	    oldWindowyPos = windowyPos;
//	    equation.increment();
//	    windowxPos = mapxPoint(Math.cos(equation.getx() - 3.141529/2.0));
//	    windowyPos = mapyPoint(Math.sin(equation.getx() - 3.141529/2.0));
//	  }
//
//
	  public void setBounds(int x, int y, int w, int h)
	  {
		  super.setBounds(x,y,w,h);
		  offScreenImage = null;
	  }
	  
	  public void setBounds(Rectangle r)
	  {
		  super.setBounds(r);
		  offScreenImage = null;
	  }
	  
	  public void setSize(Dimension d)
	  {
		  super.setSize(d);
		  offScreenImage = null;
	  }
	  
	  public void setSize(int w, int h)
	  {
		  super.setSize(w,h);
		  offScreenImage = null;
	  }
	  
	  public void setPreferredSize(Dimension d)
	  {
		  super.setPreferredSize(d);
		  offScreenImage = null;
	  }
	  
	  public void setMinimumSize(Dimension d)
	  {
		  super.setMinimumSize(d);
		  offScreenImage = null;
	  }
	  
	  public void setMaximumSize(Dimension d)
	  {
		  super.setMaximumSize(d);
		  offScreenImage = null;
	  }
	  
	  
	  /**
	   * Called as a result of any repaint() operation.
	   * Double-buffering for smooth animation!!
	   */
	  public void update(Graphics g)
	  {
	    if(offScreenImage == null) 
	    {
	      offScreenImage = createImage(size().width, size().height);
	      offScreenGraphics = offScreenImage.getGraphics();
	      clearCanvas();
	    }
	    
	    if(offScreenGraphics == null)
	    {
	      offScreenGraphics = offScreenImage.getGraphics();
	    }
	      
	    paint(offScreenGraphics);

	    // --- "paste" the offscreen image into the current
	    //     graphics object
	    g.drawImage(offScreenImage, 0, 0, this);
	  } 

	  
	  /**
	   * Use the offscreen graphics to make something happen
	   */
	  public void paint(Graphics g) 
	  {
		// --- Clear out the graphics
	    g.setColor(getBackground());
	    g.fill3DRect(0, 0, size().width, size().height, false);

	    // --- Store some precomputed variables
	    int ymax = size().height;
	    
	    // --- Draw the apparatus
	    g.setColor(Color.ORANGE);
	    
	    // --- Draw the incline and path using the length of the incline
	    //     (think of a ladder against a wall w/ fixed length)
	    int y_intercept = (int)model.getYIntercept();
	    int x_intercept = (int)model.getXIntercept();
	    Polygon p = new Polygon();
	    p.addPoint(0, ymax-INCLINE_VERTICAL_OFFSET);
	    p.addPoint(x_intercept+INCLINE_HORIZONTAL_OFFSET, ymax-INCLINE_VERTICAL_OFFSET);
	    p.addPoint(INCLINE_HORIZONTAL_OFFSET, ymax-y_intercept-INCLINE_VERTICAL_OFFSET);	    
	    p.addPoint(0, ymax-y_intercept-INCLINE_VERTICAL_OFFSET);	    
	    g.fillPolygon(p);

	    // --- Draw the friction of the incline
	    //
	    //     Change the color of the friction depending on the amount
	    if(model.getInclineFriction() < 0.11)
	    {
	    	g.setColor(Color.ORANGE);
	    }
	    else
	    {
	    	g.setColor(Color.BLACK);
	    }
	    p = new Polygon();
	    p.addPoint(INCLINE_HORIZONTAL_OFFSET, ymax-y_intercept-INCLINE_VERTICAL_OFFSET);
	    p.addPoint(INCLINE_HORIZONTAL_OFFSET+x_intercept, ymax-INCLINE_VERTICAL_OFFSET);
	    p.addPoint(INCLINE_HORIZONTAL_OFFSET+x_intercept-5,ymax-INCLINE_VERTICAL_OFFSET);
	    p.addPoint(INCLINE_HORIZONTAL_OFFSET, ymax-y_intercept+5-INCLINE_VERTICAL_OFFSET);
	    g.fillPolygon(p);
	    
	    // --- Draw the runway
    	g.setColor(Color.ORANGE);
    	int runLength_x = (int)(model.getRunLength()*Math.cos(model.getRunIncline()));
    	int runLength_y = (int)(model.getRunLength()*Math.sin(model.getRunIncline()));
    	p = new Polygon();
	    p.addPoint(x_intercept+INCLINE_HORIZONTAL_OFFSET, ymax-INCLINE_VERTICAL_OFFSET);
	    p.addPoint(x_intercept+INCLINE_HORIZONTAL_OFFSET+runLength_x, ymax-INCLINE_VERTICAL_OFFSET-runLength_y);
	    p.addPoint(size().width-1, ymax-INCLINE_VERTICAL_OFFSET-runLength_y);	    
	    p.addPoint(size().width-1, ymax-INCLINE_VERTICAL_OFFSET);	    
	    g.fillPolygon(p);

	    // --- Draw the friction of the runway
	    //
	    //     Change the color of the friction depending on the amount
	    if(model.getRunLengthFriction() < 0.11)
	    {
	    	g.setColor(Color.ORANGE);
	    }
	    else
	    {
	    	g.setColor(Color.BLACK);
	    }
	    p = new Polygon();	    
	    p.addPoint(x_intercept+INCLINE_HORIZONTAL_OFFSET, ymax-INCLINE_VERTICAL_OFFSET);
	    p.addPoint(x_intercept+INCLINE_HORIZONTAL_OFFSET+runLength_x, ymax-INCLINE_VERTICAL_OFFSET-runLength_y);
	    p.addPoint(x_intercept+INCLINE_HORIZONTAL_OFFSET+runLength_x, ymax-INCLINE_VERTICAL_OFFSET-runLength_y+5);
	    p.addPoint(x_intercept+INCLINE_HORIZONTAL_OFFSET+5, ymax-INCLINE_VERTICAL_OFFSET);
	    g.fillPolygon(p);

	    // --- Draw the base
    	g.setColor(Color.ORANGE);
	    g.fillRect(0, ymax-INCLINE_VERTICAL_OFFSET, size().width-1, size().height-1);
	    

	    // --- Draw the ball, remembering that the model's axes are different
	    //     (decreasing y in the model means increasing y in the view)
	    //     
	    if(model.getBallMass() < 1.1)
	    {
	    	g.setColor(Color.blue);
	    }
	    else
	    {
	    	g.setColor(Color.white);
	    }
	    
	    int ball_radius = (int)model.getBallRadius();
	    
	    int x_corner = (int)(model.getBallX() -
	    		            ball_radius +
	    		            INCLINE_HORIZONTAL_OFFSET);
	    int y_corner = ymax-(int)(model.getBallY() +
	    						ball_radius +
	    		                 INCLINE_VERTICAL_OFFSET);
//	    System.err.println("Ball X: " + model.getBallX() + " --> " + x_corner);
//	    System.err.println("Ball Y: " + model.getBallY() + " --> " + y_corner);
//	    System.err.println("Diam: " + ball_radius);
//	    System.err.println("OFFSET Y: " + INCLINE_VERTICAL_OFFSET);
//	    System.err.println("YMax: " + ymax);
//	    System.err.println("Size: " + getSize().toString());
	    g.fillOval(x_corner, y_corner, ball_radius*2, ball_radius*2);	    
	  }
//
//
//	  //===================================================================
//	  // Mapping functions
//	  //
//	  // Map a point in phase space onto the window.
//	  //===================================================================
//	  private int mapxPoint(double x) {
//	    return((int)((x-domainxMin)
//	           /(domainxMax-domainxMin)*(double)size().width));
//	  }
//
//	  private int mapyPoint(double y) {
//	    return((int)(-(y-domainyMax)
//	           /(domainyMax-domainyMin)*(double)size().height));
//	  }

	public void modelChangeEvent(Object o) 
	{
		this.repaint();
	}
	
	
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
	
	/**
	 * Simple test to make sure display functions correctly
	 */
	public static void main(String[] args)
	{
		Frame f = new Frame();
		InclinedPlaneCanvas c = new InclinedPlaneCanvas(new InclinedPlaneModel());
		f.add(c);
		
		f.setSize(400, 300);
		f.setVisible(true);
	}
}
